#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

package require TclExcept
package require cmdline
package require twapi

namespace eval twapi {
}

#
# Returns an keyed list with the following elements:
#   os_major_version
#   os_minor_version
#   os_build_number
#   platform - currently always NT
#   sp_major_version
#   sp_minor_version
#   suites - one or more from backoffice, blade, datacenter, enterprise, 
#            smallbusiness, smallbusiness_restricted, terminal, personal
#   system_type - workstation, server
proc twapi::get_os_info {} {
    variable windefs

    # Allocate block to hold extended information
    set vers_info [new_OSVERSIONINFOEXW]

    # First try getting extended information and if that fails, try
    # the older form
    set info_sz             276
    set extended_info_sz    284
    set have_extended_info  1
    $vers_info configure -dwOSVersionInfoSize $extended_info_sz
    if {[catch {GetVersionEx $vers_info}]} {
        $vers_info configure -dwOSVersionInfoSize $info_sz
        GetVersionEx $vers_info
        set have_extended_info 0
    }

    set osinfo(os_major_version) [$vers_info cget -dwMajorVersion]
    set osinfo(os_minor_version) [$vers_info cget -dwMinorVersion]
    set osinfo(os_build_number)  [$vers_info cget -dwBuildNumber]
    set osinfo(platform)         "NT"
    
    if {$have_extended_info} {
        set osinfo(sp_major_version) [$vers_info cget -wServicePackMajor]
        set osinfo(sp_minor_version) [$vers_info cget -wServicePackMinor]

        set osinfo(suites) [list ]
        set suites [$vers_info cget -wSuiteMask]
        foreach suite {
            backoffice blade datacenter enterprise smallbusiness
            smallbusiness_restricted terminal personal
        } {
            set def "VER_SUITE_[string toupper $suite]"
            if {$suites & $windefs($def)} {
                lappend osinfo(suites) $suite
            }
        }
        
        set system_type [$vers_info cget -wProductType]
        if {$system_type == $windefs(VER_NT_WORKSTATION)} {
            set osinfo(system_type) "workstation"
        } elseif {$system_type == $windefs(VER_NT_SERVER)} {
            set osinfo(system_type) "server"
        } elseif {$system_type == $windefs(VER_NT_DOMAIN_CONTROLLER)} {
            set osinfo(system_type) "domain_controller"
        } else {
            set osinfo(system_type) "unknown"
        }

    } else {
        # Older OS versions. Only have basic info. Figure out the rest
        # based on registry settings
        package require registry
        
        set osinfo(suites) [list ]
        set product_type [registry get "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\ProductOptions" "ProductType"]
        switch -exact -- [string toupper $product_type] {
            "WINNT" {
                set osinfo(system_type) "workstation"
            }
            "LANMANNT" {
                set osinfo(system_type) "server"
            }
            "SERVERNT" {
                set osinfo(system_type) "server"
                lappend osinfo(suites)  "enterprise"
            }
        }

        # Figure out service pack
        set sp_text [registry get "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion" "CSDVersion"]
        set sp_major 0
        regexp -nocase {Service Pack ([0-9]+)} $sp_text dummy sp_major
        set osinfo(sp_major_version) $sp_major
        set osinfo(sp_minor_version) 0; # Always 0
        
        # Terminal server?
        if {[catch {
            registry get "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\ProductOptions" "ProductSuite"
        } ts] == 0} {
            if {[string equal -nocase $ts "Terminal Server"]} {
                lappend osinfo(suites) "terminal"
            }
        }
    }
    return [array get osinfo]
}

#
# Return a text string describing the OS version and options
# If specified, osinfo should be a keyed list containing 
# data returned by get_os_info
#
proc twapi::get_os_description {} {

    array set osinfo [get_os_info]

    # Assume not terminal server
    set tserver ""

    # Version
    set osversion "$osinfo(os_major_version).$osinfo(os_minor_version)"

    # Base OS name
    if {$osinfo(os_major_version) < 5} {
        set osname "Windows NT"
        if {[string equal $osinfo(system_type) "workstation"]} {
            set systype "Workstation"
        } else {
            if {[lsearch -exact $osinfo(suites) "terminal"] >= 0} {
                set systype "Terminal Server Edition"
            } elseif {[lsearch -exact $osinfo(suites) "enterprise"] >= 0} {
                set systype "Advanced Server"
            } else {
                set systype "Server"
            }
        }
    } else {
        switch -exact -- $osversion {
            "5.0" {
                set osname "Windows 2000"
                if {[string equal $osinfo(system_type) "workstation"]} {
                    set systype "Professional"
                } else {
                    if {[lsearch -exact $osinfo(suites) "datacenter"] >= 0} {
                        set systype "Datacenter Server"
                    } elseif {[lsearch -exact $osinfo(suites) "enterprise"] >= 0} {
                        set systype "Advanced Server"
                    } else {
                        set systype "Server"
                    }
                }
            }
            "5.1" {
                set osname "Windows XP"
                if {[lsearch -exact $osinfo(suites) "personal"] >= 0} {
                    set systype "Home Edition"
                } else {
                    set systype "Professional"
                }
            }
            "5.2" {
                set osname "Windows Server 2003"
                if {[string equal $osinfo(system_type) "workstation"]} {
                    set systype "Professional"
                } else {
                    if {[lsearch -exact $osinfo(suites) "datacenter"] >= 0} {
                        set systype "Datacenter Edition"
                    } elseif {[lsearch -exact $osinfo(suites) "enterprise"] >= 0} {
                        set systype "Enterprise Edition"
                    } elseif {[lsearch -exact $osinfo(suites) "blade"] >= 0} {
                        set systype "Web Edition"
                    } else {
                        set systype "Standard Edition"
                    }
                }
            }
            default {
                # Future release - can't really name, just make something up
                set osname "Windows"
                if {[string equal $osinfo(system_type) "workstation"]} {
                    set systype "Professional"
                } else {
                    set systype "Server"
                }
            }
        }
        if {[lsearch -exact $osinfo(suites) "terminal"] >= 0} {
            set tserver " with Terminal Services"
        }
    }

    # Service pack
    if {$osinfo(sp_major_version) != 0} {
        set spver " Service Pack $osinfo(sp_major_version)"
    } else {
        set spver ""
    }
    
    return "$osname $systype ${osversion} (Build $osinfo(os_build_number))${spver}${tserver}"
}

#
# Return major minor servicepack as a quad list
proc twapi::get_os_version {} {
    if {[info exists ::twapi::_osversion]} {
        return $::twapi::_osversion
    }

    array set osinfo [get_os_info]
    set ::twapi::_osversion \
        [list $osinfo(os_major_version) $osinfo(os_minor_version) \
             $osinfo(sp_major_version) $osinfo(sp_minor_version)]
    return $::twapi::_osversion
}

# Returns true if the OS version is at least $major.$minor.$sp
proc twapi::min_os_version {major {minor 0} {spmajor 0} {spminor 0}} {
    foreach {osmajor osminor osspmajor osspminor} [twapi::get_os_version] {break}

    if {$osmajor > $major} {return 1}
    if {$osmajor < $major} {return 0}
    if {$osminor > $minor} {return 1}
    if {$osminor < $minor} {return 0}
    if {$osspmajor > $spmajor} {return 1}
    if {$osspmajor < $spmajor} {return 0}
    if {$osspminor > $spminor} {return 1}
    if {$osspminor < $spminor} {return 0}

    # Same version, ok
    return 1
}

#
# Returns proc information
#  $processor should be processor number or "" for "total"
proc twapi::get_processor_info {processor args} {
    # Note the PDH options match those of 
    # twapi::get_processor_perf_counter_paths
    set pdh_opts {
        dpcutilization
        interruptutilization
        privilegedutilization
        processorutilization
        userutilization
        apcbypassrate
        dpcbypassrate
        dpcrate
        dpcqueuerate
        interruptrate
    }

    array set opts [cmdline::getoptions args \
                        [concat [list all \
                                     [list interval.arg 100]] \
                             $pdh_opts]]

    set result [list ]

    # Now retrieve each PDH counter
    set requested_opts [list ]
    foreach pdh_opt $pdh_opts {
        if {$opts(all) || $opts($pdh_opt)} {
            lappend requested_opts "-$pdh_opt"
        }
    }
    
    set counter_list [eval [list get_perf_processor_counter_paths $processor] \
                          $requested_opts -refresh]

    set results [list ]
    foreach {opt processor value} [get_perf_values_from_metacounter_info $counter_list -interval $opts(interval)] {
        lappend results -$opt $value
    }
    return $results
}


#
# Get the netbios name
proc twapi::get_computer_netbios_name {} {
    return [GetComputerName]
}
