#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

# TBD - allow access rights to be specified symbolically using procs
# from security.tcl

package require TclExcept
package require cmdline
package require twapi

namespace eval twapi {
}

#
# Get my process id
proc twapi::get_current_process_id {} {
    return [::pid]
}

#
# Get my thread id
proc twapi::get_current_thread_id {} {
    return [GetCurrentThreadId]
}

# Return list of process ids
# Note if -path or -name is specified, then processes for which this
# information cannot be obtained are skipped
proc twapi::get_process_ids {args} {

    set save_args $args;                # Need to pass to process_exists
    array set opts [cmdline::getoptions args { path.arg name.arg glob}]

    if {[string length $opts(path)] && [string length $opts(name)]} {
        error "Options -path and -name are mutually exclusive"
    }

    set pids [twapi::EnumProcesses]
    if {[string length $opts(path)] == 0 && [string length $opts(name)] == 0} {
        # Return all pids
        return $pids
    }
    
    # Only want pids with a specific path or name

    # Lazy - just assume process_exists has the same options and pass them
    # to it
    set process_pids [list ]
    foreach pid $pids {
        if {[eval [list process_exists $pid] $save_args] > 0} {
            lappend process_pids $pid
        }
    }
    return $process_pids
}


# Return list of modules handles for a process
proc twapi::get_process_modules {pid args} {
    variable windefs

    array set opts [cmdline::getoptions args {handle name path imagedata all}]

    if {$opts(all)} {
        foreach opt {handle name path imagedata} {
            set opts($opt) 1
        }
    }
    set noopts [expr {($opts(name) || $opts(path) || $opts(imagedata) || $opts(handle)) == 0}]

    set privs [expr {$windefs(PROCESS_QUERY_INFORMATION) | $windefs(PROCESS_VM_READ)}]
    set hpid [OpenProcess $privs 0 $pid]
    set results [list ]
    try {
        foreach module [EnumProcessModules $hpid] {
            if {$noopts} {
                lappend results $module
                continue
            }
            set module_data [list ]
            if {$opts(handle)} {
                lappend module_data -handle $module
            }
            if {$opts(name)} {
                if {[catch {GetModuleBaseName $hpid $module} name]} {
                    set name ""
                }
                lappend module_data -name $name
            }
           if {$opts(path)} {
                if {[catch {GetModuleFileNameEx $hpid $module} path]} {
                    set path ""
                }
               lappend module_data -path [_normalize_path $path]
            }
            if {$opts(imagedata)} {
                if {[catch {GetModuleInformation $hpid $module} imagedata]} {
                    set base ""
                    set size ""
                    set entry ""
                } else {
                    array set temp $imagedata
                    set base $temp(lpBaseOfDll)
                    set size $temp(SizeOfImage)
                    set entry $temp(EntryPoint)
                }                
                lappend module_data -imagedata [list $base $size $entry]
            }
            lappend results $module_data
        }
    } finally {
        CloseHandle $hpid
    }
    return $results
}



#
# Get the path of a process
proc twapi::get_process_path {pid args} {
    return [eval [list twapi::_get_process_name_path_helper $pid path] $args]
}


#
# Get the path of a process
proc twapi::get_process_name {pid args} {
    return [eval [list twapi::_get_process_name_path_helper $pid name] $args]
}


# Return list of device drivers
proc twapi::get_device_drivers {args} {
    variable windefs

    array set opts [cmdline::getoptions args {name path base all}]

    set results [list ]
    foreach module [EnumDeviceDrivers] {
        catch {unset module_data}
        if {$opts(base) || $opts(all)} {
            set module_data [list -base $module]
        }
        if {$opts(name) || $opts(all)} {
            if {[catch {GetDeviceDriverBaseName $module} name]} {
                    set name ""
            }
            lappend module_data -name $name
        }
        if {$opts(path) || $opts(all)} {
            if {[catch {GetDeviceDriverFileName $module} path]} {
                set path ""
            }
            lappend module_data -path [_normalize_path $path]
        }
        if {[info exists module_data]} {
            lappend results $module_data
        }
    }

    return $results
}

#
# Kill a process 
# Returns 1 if process was ended, 0 if not ended within timeout
# 
proc twapi::end_process {pid args} {

    array set opts [cmdline::getoptions args {
        {exitcode.arg 1}
        {force    0}
        {wait.arg 0}
    }]
    
    set process_path [get_process_path $pid]

    # First try to close nicely
    set toplevels [get_toplevel_windows -pid $pid]
    if {[llength $toplevels]} {
        # Try and close by sending them a message. WM_CLOSE is 0x10
        foreach toplevel $toplevels {
            # Send a message but come back right away
            if {0} {
                catch {PostMessage $toplevel 0x10 0 0}
            } else {
                catch {SendNotifyMessage $toplevel 0x10 0 0}
            }
        }

        # Wait for the specified time to verify process has gone away
        set gone [twapi::wait {process_exists $pid -path $process_path} 0 $opts(wait)]
        if {$gone || ! $opts(force)} {
            # Succeeded or do not want to force a kill
            return $gone
        }
        
        # Only wait 10 ms since we have already waited above
        if {$opts(wait)} {
            set opts(wait) 10
        }
    }

    set hpid [OpenProcess $twapi::windefs(PROCESS_TERMINATE) 0 $pid]
    try {
        TerminateProcess $hpid $opts(exitcode)
    } finally {
        CloseHandle $hpid
    }

    return [twapi::wait {process_exists $pid -path $process_path} 0 $opts(wait)]
}


# Check if the given process exists
# 0 - does not exist or exists but paths/names do not match, 
# 1 - exists and matches path (or no -path or -name specified)
# -1 - exists but do not know path and cannot compare
proc twapi::process_exists {pid args} {
    array set opts [cmdline::getoptions args { path.arg name.arg glob}]

    if {[string length $opts(path)] && [string length $opts(name)]} {
        error "Options -path and -name are mutually exclusive"
    }

    if {[string length $opts(path)]} {
        set get_proc get_process_path
    } else {
        set get_proc get_process_name
    }

    set process_path [$get_proc $pid -noexist "" -noaccess "(unknown)"]
    if {[string length $process_path] == 0} {
        # No such process
        return 0
    }
    # Process with this pid exists

    if {[string length $opts(path)] == 0 && [string length $opts(name)] == 0} {
        # Don't care about the path or name
        return 1
    }

    # The process exists, but paths/names have to match
    if {[string equal $process_path "(unknown)"]} {
        # Exists but cannot check path/name
        return -1
    }

    if {$opts(glob)} {
        set string_cmd match
    } else {
        set string_cmd equal
    }

    if {[string length $opts(path)]} {
        # Note we do not use file normalize here since that will tack on
        # absolute paths which we do not want for glob matching

        # We use [file join ] to convert \ to / to avoid special
        # interpretation of \ in string match command
        return [string $string_cmd -nocase [file join $opts(path)] [file join $process_path]]
    } else {
        return [string $string_cmd -nocase $opts(name) $process_path]
    }
}

#
# Get the parent process of a thread. Return "" if no such thread
proc twapi::get_thread_parent_process_id {tid} {
    set pid_paths [get_perf_thread_counter_paths $tid -pid]
    if {[llength $pid_paths] == 0} {
        return ""
    }

    if {[get_counter_path_value [lindex [lindex $pid_paths 0] 3] -var pid]} {
        return $pid
    } else {
        return ""
    }
}

#
# Get the thread ids belonging to a process
proc twapi::get_process_thread_ids {pid} {
    set counter_paths [get_perf_instance_counter_paths \
                           "Thread" [list "ID Thread"] "ID Process" \
                           [list $pid] -refresh]
    set tids [list ]
    foreach counter_path $counter_paths {
        if {[get_counter_path_value $counter_path -var tid]} {
            lappend tids $tid
        }
    }
    return $tids
}


#
# Get process information 
proc twapi::get_process_info {pid args} {
    return [lindex [eval [list get_multiple_process_info [list $pid]] $args] 1]
}


#
# Get multiple process information 
proc twapi::get_multiple_process_info {pids args} {

    # Note the PDH options match those of twapi::get_process_perf_counter_paths
    set pdh_opts {
        privilegedutilization
        processorutilization
        userutilization
        parent
        elapsedtime
        handlecount
        pid
        iodatabytesrate
        iodataopsrate
        iootherbytesrate
        iootheropsrate
        ioreadbytesrate
        ioreadopsrate
        iowritebytesrate
        iowriteopsrate
        pagefaultrate
        pagefilebytes
        pagefilebytespeak
        poolnonpagedbytes
        poolpagedbytes
        basepriority
        privatebytes
        threadcount
        virtualbytes
        virtualbytespeak
        workingset
        workingsetpeak
    }

    array set opts [cmdline::getoptions args \
                        [concat [list all \
                                     path \
                                     name \
                                     tids \
                                     toplevels \
                                     [list interval.arg 100]] \
                             $pdh_opts]]

    array set results {}

    foreach pid $pids {
        set result [list ]

        if {$opts(all) || $opts(name)} {
            lappend result -name [get_process_name $pid]
        }
        
        if {$opts(all) || $opts(path)} {
            lappend result -path [get_process_path $pid]
        }
        
        if {$opts(all) || $opts(tids)} {
            lappend result -tids [get_process_thread_ids $pid]
        }
        
        if {$opts(all) || $opts(toplevels)} {
            lappend result -toplevels [get_toplevel_windows -pid $pid]
        }
        # Now retrieve each PDH counter
        set requested_opts [list ]
        foreach pdh_opt $pdh_opts {
            if {$opts(all) || $opts($pdh_opt)} {
                lappend requested_opts "-$pdh_opt"
            }
        }
    
        set results($pid) $result
    }

    set counters [eval [list get_perf_process_counter_paths $pids] \
                      $requested_opts -refresh]
    foreach {opt pid val} [get_perf_values_from_metacounter_info $counters -interval $opts(interval)] {
        lappend results($pid) $opt $val
    }
    return [array get results]

    set counter_list [list ]
    foreach {pid pid_counters} [array get counters] {
        foreach pid_counter $pid_counters {
            set pid_counter [lreplace $pid_counter 0 0 [list $pid [lindex $pid_counter 0]]]
            lappend counter_list $pid_counter
        }
        #set counter_list [concat $counter_list $pid_counters]
    }
    foreach {key val} [get_perf_values_from_metacounter_info $counter_list -interval $opts(interval)] {
        foreach {pid opt} $key {break}
        lappend results($pid) $opt $val
    }

    return [array get results]
}


#
# Get thread information 
# TBD - add info from GetGUIThreadInfo
proc twapi::get_thread_info {tid args} {
    # Note the PDH options match those of twapi::get_thread_perf_counter_paths
    set pdh_opts {
        privilegedutilization 
        processorutilization
        userutilization
        contextswitches
        elapsedtime
        pid
        tid
        basepriority
        priority
        startaddress
        state
        waitreason
    }

    array set opts [cmdline::getoptions args \
                        [concat [list all [list interval.arg 100]] $pdh_opts]]

    # Now retrieve each PDH counter
    set requested_opts [list ]
    foreach pdh_opt $pdh_opts {
        if {$opts(all) || $opts($pdh_opt)} {
            lappend requested_opts "-$pdh_opt"
        }
    }
    
    set counter_list [eval [list get_perf_thread_counter_paths [list $tid]] \
                          $requested_opts -refresh]
    set results [list ]
    foreach {opt tid value} [get_perf_values_from_metacounter_info $counter_list -interval $opts(interval)] {
        lappend results $opt $value
    }
    return $results
}




#
# Utility procedures
#


# Convert file names by substituting \SystemRoot and \??\ sequences
proc twapi::_normalize_path {path} {
    global env
    
    # Get rid of \??\ prefixes
    regsub {^[\\/]\?\?[\\/](.*)} $path {\1} path

    # Replace leading \SystemRoot with real system root
    catch {set systemroot $env(WINDIR)}
    catch {set systemroot $env(SYSTEMROOT)}
    regsub -nocase {^[\\/]systemroot([\\/].*)} $path "${systemroot}\\1" path

    # If not a full path, make it a full path

    return $path
}

#
# Get the path of a process
proc twapi::_get_process_name_path_helper {pid {type name} args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        {noexist.arg "(no such process)"}
        {noaccess.arg "(unknown)"}
    }]

    if {![string is integer $pid]} {
        error "Invalid non-numeric pid $pid"
    }
    # System pid is 8 on Win2K, 4 on XP. In both cases, the other is not used
    # for a real process so don't need a version check here
    if {$pid == 8 || $pid == 4} {
        return "System"
    }
    if {$pid == 0} {
        return "System Idle Process"
    }

    set privs [expr {$windefs(PROCESS_QUERY_INFORMATION) | $windefs(PROCESS_VM_READ)}]

    try {
        set hpid [OpenProcess $privs 0 $pid]
    } catch msg {
        default {
            foreach {subsystem code msg} $::errorCode {break}
            if {[string equal -nocase $subsystem "WINDOWS"]} {
                if {$code == 87} {
                    return $opts(noexist)
                } elseif {$code == 5} {
                    return $opts(noaccess)
                } else {
                    throw
                }
            } else {
                throw
            }
        }
    }
    try {
        set module [lindex [EnumProcessModules $hpid] 0]
        if {[string equal $type "name"]} {
            set path [GetModuleBaseName $hpid $module]
        } else {
            set path [_normalize_path [GetModuleFileNameEx $hpid $module]]
        }
    } finally {
        CloseHandle $hpid
    }
    return $path
}
