#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

# TBD - allow SID and account name to be used interchangeably in various
# functions
# TBD - ditto for LUID v/s privilege names

package require TclExcept
package require cmdline
package require twapi

namespace eval twapi {
    # Map of Sid integer type to Sid type name
    array set sid_type_names {
        1 user 
        2 group
        3 domain 
        4 alias 
        5 wellknowngroup
        6 deletedaccount
        7 invalid
        8 unknown
        9 computer
    }
}

#
# Helper for lookup_account_name{sid,name}
proc twapi::_lookup_account {func account args} {
    if {$func == "LookupAccountName"} {
        set lookup sid
    } else {
        set lookup name
    }
    array set opts [cmdline::getoptions args \
                        [list all \
                             $lookup \
                             domain \
                             type \
                             system.arg]]

    foreach "$lookup domain type" [$func $opts(system) $account] break

    set result [list ]
    if {$opts(all) || $opts(domain)} {
        lappend result -domain $domain
    }
    if {$opts(all) || $opts(type)} {
        lappend result -type $twapi::sid_type_names($type)
    }

    if {$opts(all) || $opts($lookup)} {
        lappend result -$lookup [set $lookup]
    }

    # If no options specified, only return the sid/name
    if {[llength $result] == 0} {
        return [set $lookup]
    }

    return $result
}

# Returns the sid, domain and type for an account
proc twapi::lookup_account_name {name args} {
    return [eval [list _lookup_account LookupAccountName $name] $args]
}


# Returns the name, domain and type for an account
proc twapi::lookup_account_sid {sid args} {
    return [eval [list _lookup_account LookupAccountSid $sid] $args]
}

#
# Returns the sid for a account - may be given as a SID or name
proc twapi::map_account_to_sid {account args} {
    array set opts [cmdline::getoptions args [list system.arg]]

    if {[is_valid_sid_syntax $account]} {
        return $account
    } else {
        return [lookup_account_name $account -system $opts(system)]
    }
}


#
# Returns the name for a account - may be given as a SID or name
proc twapi::map_account_to_name {account args} {
    array set opts [cmdline::getoptions args [list system.arg]]

    if {[is_valid_sid_syntax $account]} {
        return [lookup_account_sid $account -system $opts(system)]
    } else {
        # Verify whether a valid account by mapping to an sid
        if {[catch {map_account_to_sid $account -system $opts(system)}]} {
            error "Unknown account '$account'"
        } 
        return $account
    }
}


#
# Verify that the given sid is valid. This is purely a syntactic check
proc twapi::is_valid_sid_syntax sid {
    try {
        set result [IsValidSid $sid]
    } catch msg {
        default {
            if {[lindex $::errorCode 0] == "WINDOWS"
                && [lindex $::errorCode 1] == 1337} {
                set result 0
            } else {
                throw
            }
        }
    }
    return $result
}

#
# Returns token for the process with pid $pid
# If $pid=="", the current process is assumed
proc twapi::open_process_token {pid access} {
    variable windefs

    set access [_access_rights_to_mask $access]

    # If "Win2K all access" specified, modify for NT 4.0 else we get an error
    if {($access == $windefs(TOKEN_ALL_ACCESS_WIN2K))
        && ([lindex [get_os_version] 0] == 4)} {
        set access $windefs(TOKEN_ALL_ACCESS_WINNT)
    }

    # Get a handle for the process
    if {[string length $pid]} {
        set ph [OpenProcess $windefs(PROCESS_QUERY_INFORMATION) 0 $pid]
    } else {
        set ph [GetCurrentProcess]
    }
    try {
        # Get a token for the process
        set ptok [OpenProcessToken $ph $access]
    } finally {
        # Close handle only if we did an OpenProcess
        if {[string length $pid]} {
            CloseHandle $ph
        }
    }

    return $ptok
}


#
# Returns token for the thread with tid $tid. If $tid is "", current thread
# is assumed
proc twapi::open_thread_token {tid access} {
    variable windefs

    set access [_access_rights_to_mask $access]

    # If "Win2K all access" specified, modify for NT 4.0 else we get an error
    if {($access == $windefs(TOKEN_ALL_ACCESS_WIN2K))
        && ([lindex [get_os_version] 0] == 4)} {
        set access $windefs(TOKEN_ALL_ACCESS_WINNT)
    }

    # Get a handle for the thread
    if {[string length $tid]} {
        set th [OpenThread $tid]
    } else {
        set th [GetCurrentThread]
    }
    try {
        # Get a token for the process
        set ttok [OpenThreadToken $ph $opts(self) $access]
    } finally {
        CloseHandle $ph
    }

    return $ttok
}

#
# Close a token
proc twapi::close_token {tok} {
    CloseHandle $tok
}

#
# Get the user account associated with a token
proc twapi::get_token_user {tok args} {
    array set opts [cmdline::getoptions args [list name]]
    set user [lindex [GetTokenInformation $tok $twapi::windefs(TokenUser)] 0]
    if {$opts(name)} {
        set user [lookup_account_sid $user]
    }
    return $user
}

#
# Get the groups associated with a token
proc twapi::get_token_groups {tok args} {
    array set opts [cmdline::getoptions args [list name]]

    set groups [list ]
    foreach {group} [GetTokenInformation $tok $twapi::windefs(TokenGroups)] {
        set group [lindex $group 0]
        if {$opts(name)} {
            set group [lookup_account_sid $group]
        }
        lappend groups $group
    }

    return $groups
}

#
# Get the groups associated with a token along with their attributes
# These are returned as a flat list of the form "sid attrlist sid attrlist..."
# where the attrlist is a list of attributes
proc twapi::get_token_group_sids_and_attrs {tok} {
    variable windefs 

    set sids_and_attrs [list ]
    foreach {group} [GetTokenInformation $tok $windefs(TokenGroups)] {
        foreach {sid attr} $group break
        set attr_list {enabled enabled_by_default logon_id
            mandatory owner resource use_for_deny_only}
        lappend sids_and_attrs $sid [_map_token_attr $attr $attr_list SE_GROUP]
    }

    return $sids_and_attrs
}

# Get list of privileges that are currently enabled for the token
# If -all is specified, gets all privileges
proc twapi::get_token_privileges {tok args} {
    variable windefs

    set all [expr {[lsearch -exact $args -all] >= 0}]

    set privs [list ]
    foreach {item} [GetTokenInformation $tok $windefs(TokenPrivileges)] {
        if {$all || [expr {[lindex $item 1] & $windefs(SE_PRIVILEGE_ENABLED)}]} {
            lappend privs [map_luid_to_privilege [lindex $item 0] -mapunknown]
        }
    }
    return $privs
}

#
# Return true if the token has the given privilege
# $args is concat before checking
proc twapi::check_enabled_privileges {tok privlist args} {
    set all_required [expr {[lsearch -exact $args "-any"] < 0}]

    if {0} {
        We now call the PrivilegeCheck instead. Not sure it matters
        This code also does not handle -any option
        foreach priv $privlist {
            if {[expr {
                       [lsearch -exact [get_token_privileges $tok] $priv] < 0
                   }]} {
                return 0
            }
        }
        return 1
    } else {
        set luid_attr_list [list ]
        foreach priv $privlist {
            lappend luid_attr_list [list [map_privilege_to_luid $priv] 0]
        }
        return [Twapi_PrivilegeCheck $tok $luid_attr_list $all_required]
    }
}


#
# 

# Get the privilege associated with a token and their attributes
proc twapi::get_token_privileges_and_attrs {tok} {
    set privs_and_attrs [list ]
    foreach priv [GetTokenInformation $tok $twapi::windefs(TokenPrivileges)] {
        foreach {luid attr} $priv break
        set attr_list {enabled enabled_by_default used_for_access}
        lappend privs_and_attrs [map_luid_to_privilege $luid -mapunknown] \
            [_map_token_attr $attr $attr_list SE_PRIVILEGE]
    }

    return $privs_and_attrs

}


#
# Get the sid that will be used as the owner for objects created using this
# token. Returns name instead of sid if -name options specified
proc twapi::get_token_owner {tok args} {
    return [ _get_token_sid_field $tok TokenOwner $args]
}


#
# Get the sid that will be used as the primary group for objects created using
# this token. Returns name instead of sid if -name options specified
proc twapi::get_token_primary_group {tok args} {
    return [ _get_token_sid_field $tok TokenPrimaryGroup $args]
}


#
# Return the source of an access token
proc twapi::get_token_source {tok} {
    return [GetTokenInformation $tok $twapi::windefs(TokenSource)]
}


#
# Return the token type of an access token
proc twapi::get_token_type {tok} {
    if {[GetTokenInformation $tok $twapi::windefs(TokenType)]} {
        return "primary"
    } else {
        return "impersonation"
    }
}

#
# Return the token type of an access token
proc twapi::get_token_impersonation_level {tok} {
    set ilevel [GetTokenInformation $tok \
                    $twapi::windefs(TokenImpersonationLevel)]
    switch -exact -- $ilevel {
        0 { return "anonymous" }
        1 { return "identification" }
        2 { return "impersonation" }
        3 { return "delegation" }
        default { return $ilevel }
    }
}

#
# Return token statistics
proc twapi::get_token_statistics {tok} {
    set stats [list ]
    set labels {luid authluid expiration type impersonationlevel
        dynamiccharged dynamicavailable groupcount
        privilegecount modificationluid}
    set statinfo [GetTokenInformation $tok $twapi::windefs(TokenStatistics)]
    foreach label $labels  val $statinfo {
        lappend stats $label $val
    }
    return $stats
}


#
# Enable the privilege state of a token. Generates an error if
# the specified privileges do not exist in the token (either
# disabled or enabled), or cannot be adjusted
proc twapi::enable_token_privileges {tok privs} {
    variable windefs

    set luid_attrs [list]
    foreach priv $privs {
        lappend luid_attrs [list [map_privilege_to_luid $priv] $windefs(SE_PRIVILEGE_ENABLED)]
    }
    Twapi_AdjustTokenPrivileges $tok 0 $luid_attrs
}

#
# Disable the privilege state of a token. Generates an error if
# the specified privileges do not exist in the token (either
# disabled or enabled), or cannot be adjusted
proc twapi::disable_token_privileges {tok privs} {
    set luid_attrs [list]
    foreach priv $privs {
        lappend luid_attrs [list [map_privilege_to_luid $priv] 0]
    }
    Twapi_AdjustTokenPrivileges $tok 0 $luid_attrs
}

#
# Disable all privs in a token
proc twapi::disable_all_token_privileges {tok} {
    Twapi_AdjustTokenPrivileges $tok 1 [list ]
}


#
# Get list of users on a system
proc twapi::get_users {args} {
    array set opts [cmdline::getoptions args {system.arg}]
    return [Twapi_NetUserEnum $opts(system) 0]
}

#
# Add a new user account
proc twapi::new_user {username args} {
    array set opts [cmdline::getoptions args {
        system.arg
        password.arg
        comment.arg
        {priv.arg "user"}
        home_dir.arg
        script_path.arg
    }]

    switch -exact -- $opts(priv) {
        guest {set priv 0}
        user  {set priv 1}
        admin {set priv 2}
        default {
            error "Invalid value '$opts(priv)' specified for -priv. Must be one of 'guest', 'user' or 'admin'"
        }
    }

    NetUserAdd $opts(system) $username $opts(password) $priv \
        $opts(home_dir) $opts(comment) 0 $opts(script_path)
}


#
# Delete a user account
proc twapi::delete_user {username args} {
    eval set [cmdline::getoptions args {
        system.arg
    }]

    NetUserDel $system $username
}


#
# Define various functions to set various user account fields
foreach field {name password home_dir comment script_path full_name country_code profile home_dir_drive} {
    proc twapi::set_user_$field {username fieldval args} "
        array set opts \[cmdline::getoptions args {
            system.arg
        }\]
        Twapi_NetUserSetInfo_$field \$opts(system) \$username \$password"
}

#
# Set user privilege level
proc twapi::set_user_priv_level {username priv_level args} {
    eval set [cmdline::getoptions args {
        system.arg
    }]

    switch -exact -- $priv_level {
        guest {set priv 0}
        user  {set priv 1}
        admin {set priv 2}
        default {
            error "Invalid privilege level value '$priv_level' specified. Must be one of 'guest', 'user' or 'admin'"
        }
    }
    Twapi_NetUserSetInfo_priv $system $username $priv
}

#
# Set account expiry time
proc twapi::set_user_expiration {username time args} {
    eval set [cmdline::getoptions args {
        system.arg
    }]

    if {[string equal $time "never"]} {
        set time -1
    } else {
        set time [clock scan $time]
    }

    Twapi_NetUserSetInfo_acct_expires $system $username $time
}

#
# Unlock a user account
proc twapi::unlock_user {username args} {
    eval [list _change_usri3_flags $username $twapi::windefs(UF_LOCKOUT) 0] $args
}

#
# Enable a user account
proc twapi::enable_user {username args} {
    eval [list _change_usri3_flags $username $twapi::windefs(UF_ACCOUNTDISABLE) 0] $args
}

#
# Disable a user account
proc twapi::disable_user {username args} {
    eval [list _change_usri3_flags $username $twapi::windefs(UF_ACCOUNTDISABLE) 1] $args
}



#
# Return the specified fields for a user account
proc twapi::get_user_account_info {account args} {
    variable windefs

    # Define each option, the corresponding field, and the 
    # information level at which it is returned
    array set fields {
        comment {usri3_comment 1}
        password_expired {usri3_password_expired 3}
        full_name {usri3_full_name 2}
        parms {usri3_parms 2}
        units_per_week {usri3_units_per_week 2}
        primary_group_id {usri3_primary_group_id 3}
        status {usri3_flags 1}
        logon_server {usri3_logon_server 2}
        country_code {usri3_country_code 2}
        home_dir {usri3_home_dir 1}
        password_age {usri3_password_age 1}
        home_dir_drive {usri3_home_dir_drive 3}
        num_logons {usri3_num_logons 2}
        acct_expires {usri3_acct_expires 2}
        last_logon {usri3_last_logon 2}
        user_id {usri3_user_id 3}
        usr_comment {usri3_usr_comment 2}
        bad_pw_count {usri3_bad_pw_count 2}
        code_page {usri3_code_page 2}
        logon_hours {usri3_logon_hours 2}
        workstations {usri3_workstations 2}
        last_logoff {usri3_last_logoff 2}
        name {usri3_name 0}
        script_path {usri3_script_path 1}
        priv {usri3_priv 1}
        profile {usri3_profile 3}
        max_storage {usri3_max_storage 2}
    }
    # Left out - auth_flags {usri3_auth_flags 2}
    # Left out (always returned as NULL) - password {usri3_password 1}


    array set opts [cmdline::getoptions args \
                        [concat [array names fields] \
                             [list local_groups global_groups system.arg all]]]

    if {$opts(all)} {
        foreach field [array names fields] {
            set opts($field) 1
        }
        set opts(local_groups) 1
        set opts(global_groups) 1
    }

    # Based on specified fields, figure out what level info to ask for
    set level 0
    foreach {field fielddata} [array get fields] {
        if {[lindex $fielddata 1] > $level} {
            set level [lindex $fielddata 1]
        }
    }
    
    array set data [NetUserGetInfo $opts(system) $account $level]

    # Extract the requested data
    array set result [list ]
    foreach {field fielddata} [array get fields] {
        if {$opts($field)} {
            set result($field) $data([lindex $fielddata 0])
        }
    }

    # Map internal values to more friendly formats
    if {$opts(status)} {
        if {$result(status) & $windefs(UF_ACCOUNTDISABLE)} {
            set result(status) "disabled"
        } elseif {$result(status) & $windefs(UF_LOCKOUT)} {
            set result(status) "locked"
        } else {
            set result(status) "enabled"
        }
    }

    if {[info exists result(logon_hours)]} {
        binary scan $result(logon_hours) b* result(logon_hours)
    }

    foreach time_field {acct_expires last_logon last_logoff} {
        if {[info exists result($time_field)]} {
            if {$result($time_field) == -1} {
                set result($time_field) "never"
            } elseif {$result($time_field) == 0} {
                set result($time_field) "unknown"
            } else {
                set result($time_field) [clock format $result($time_field) -gmt 1]
            }
        }
    }
    
    if {[info exists result(priv)]} {
        switch -exact -- [expr {$result(priv) & 3}] {
            0 { set result(priv) "guest" }
            1 { set result(priv) "user" }
            2 { set result(priv) "admin" }
        }
    }

    if {$opts(local_groups)} {
        set result(local_groups) [NetUserGetLocalGroups $opts(system) $account]
    }

    if {$opts(global_groups)} {
        set result(global_groups) [NetUserGetGroups $opts(system) $account]
    }

    return [get_array_as_options result]
}


#
# Set the specified fields for a user account
proc twapi::set_user_account_info {account args} {
    variable windefs

    set notspecified "3kjafnq2or2034r12"; # Some junk

    # Define each option, the corresponding field, and the 
    # information level at which it is returned
    array set opts [cmdline::getoptions args "
        system.arg
        [list comment.arg $notspecified]
        [list full_name.arg $notspecified]
        [list country_code.arg $notspecified]
        [list home_dir.arg $notspecified]
        [list home_dir_drive.arg $notspecified]
        [list acct_expires.arg $notspecified]
        [list name.arg $notspecified]
        [list script_path.arg $notspecified]
        [list priv.arg $notspecified]
        [list profile.arg $notspecified]"]

    if {[string compare $opts(comment) $notspecified]} {
        set_user_comment $account $opts(comment) -system $opts(system)
    }

    if {[string compare $opts(full_name) $notspecified]} {
        set_user_full_name $account $opts(full_name) -system $opts(system)
    }

    if {[string compare $opts(country_code) $notspecified]} {
        set_user_country_code $account $opts(country_code) -system $opts(system)
    }

    if {[string compare $opts(home_dir) $notspecified]} {
        set_user_home_dir $account $opts(home_dir) -system $opts(system)
    }

    if {[string compare $opts(home_dir_drive) $notspecified]} {
        set_user_home_dir_drive $account $opts(home_dir_drive) -system $opts(system)
    }

    if {[string compare $opts(acct_expires) $notspecified]} {
        set_user_expiration $account $opts(acct_expires) -system $opts(system)
    }

    if {[string compare $opts(name) $notspecified]} {
        set_user_name $account $opts(name) -system $opts(system)
    }

    if {[string compare $opts(script_path) $notspecified]} {
        set_user_script_path $account $opts(script_path) -system $opts(system)
    }

    if {[string compare $opts(priv) $notspecified]} {
        set_user_priv_level $account $opts(priv) -system $opts(system)
    }

    if {[string compare $opts(profile) $notspecified]} {
        set_user_profile $account $opts(profile) -system $opts(system)
    }

}
                    
#
# Get list of global groups on a system
proc twapi::get_global_groups {args} {
    array set opts [cmdline::getoptions args {system.arg}]
    return [NetGroupEnum $opts(system)]
}


#
# Get list of local groups on a system
proc twapi::get_local_groups {args} {
    array set opts [cmdline::getoptions args {system.arg}]
    return [NetLocalGroupEnum $opts(system)]
}


#
# Create a new global group
proc twapi::new_global_group {grpname args} {
    array set opts [cmdline::getoptions args {
        system.arg
        comment.arg
    }]

    NetGroupAdd $opts(system) $grpname $opts(comment)
}


#
# Create a new local group
proc twapi::new_local_group {grpname args} {
    array set opts [cmdline::getoptions args {
        system.arg
        comment.arg
    }]

    NetLocalGroupAdd $opts(system) $grpname $opts(comment)
}


#
# Delete a global group
proc twapi::delete_global_group {grpname args} {
    eval set [cmdline::getoptions args {system.arg}]

    NetGroupDel $opts(system) $grpname
}

#
# Delete a local group
proc twapi::delete_local_group {grpname args} {
    array set opts [cmdline::getoptions args {system.arg}]

    NetLocalGroupDel $opts(system) $grpname
}


#
# Enumerate members of a global group
proc twapi::get_global_group_members {grpname args} {
    array set opts [cmdline::getoptions args {system.arg}]

    NetGroupGetUsers $opts(system) $grpname
}

#
# Enumerate members of a local group
proc twapi::get_local_group_members {grpname args} {
    array set opts [cmdline::getoptions args {system.arg}]

    NetLocalGroupGetMembers $opts(system) $grpname
}

#
# Add a user to a global group
proc twapi::add_user_to_global_group {grpname username args} {
    eval set [cmdline::getoptions args {system.arg}]

    NetGroupAddUser $system $grpname $username
}


#
# Add a user to a local group
proc twapi::add_member_to_local_group {grpname username args} {
    eval set [cmdline::getoptions args {system.arg}]

    Twapi_NetLocalGroupAddMember $system $grpname $username
}


# Remove a user from a global group
proc twapi::remove_user_from_global_group {grpname username args} {
    eval set [cmdline::getoptions args {system.arg}]

    NetGroupDelUser $system $grpname $username
}


# Remove a user from a local group
proc twapi::remove_member_from_local_group {grpname username args} {
    eval set [cmdline::getoptions args {system.arg}]

    Twapi_NetLocalGroupDelMember $system $grpname $username
}


#
# Map a privilege given as a LUID
proc twapi::map_luid_to_privilege {luid args} {
    
    array set opts [cmdline::getoptions args [list systemname.arg mapunknown]]

    # luid may in fact be a privilege name. Check for this
    if {[is_valid_luid_syntax $luid]} {
        try {
            set name [LookupPrivilegeName $opts(systemname) $luid]
        } catch msg {
            default {
                if {[lindex $::errorCode 0] == "WINDOWS"
                    && [lindex $::errorCode 1] == 1313
                    && $opts(mapunknown]} {
                    
                    set name "Privilege-$luid"
                } else {
                    throw
                }
            }
        }
    } else {
        # Not a valid LUID syntax. Check if it's a privilege name
        if {[catch {map_privilege_to_luid $luid -system $opts(system)}]} {
            error "Invalid LUID '$luid'"
        }
        return $luid;                   # $luid is itself a priv name
    }

    return $name
}


#
# Map a privilege to a LUID
proc twapi::map_privilege_to_luid {priv args} {
    array set opts [cmdline::getoptions args [list system.arg]]

    # First check for privilege names we might have generated
    if {[string match "Privilege-*" $priv]} {
        set priv [string range $priv 10 end]
    }

    # If already a LUID format, return as is
    if {[is_valid_luid_syntax $priv]} {
        return $priv
    }
    return [LookupPrivilegeValue $opts(system) $priv]
}


#
# Return 1/0 if in LUID format
proc twapi::is_valid_luid_syntax {luid} {
    return [regexp {[[:xdigit:]]{8}-[[:xdigit:]]{8}} $luid]
}


################################################################
# Functions related to ACE's and ACL's

#
# Create a new ACE
proc twapi::new_ace {type account rights args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        self.arg 1
        recursecontainers 0
        recurseobjects 0
        recurseonelevelonly 0
    }]

    set sid [map_account_to_sid $account]

    set access_mask [_access_rights_to_mask $rights]

    switch -exact -- $type {
        allow -
        deny  -
        audit {
            set typecode [_ace_type_symbol_to_code $type]
        }
        default {
            error "Invalid or unsupported ACE type '$type'"
        }
    }

    set inherit_flags 0
    if {! $opts(self)} {
        setbits inherit_flags $windefs(INHERIT_ONLY_ACE)
    }

    if {$opts(recursecontainers)} {
        setbits inherit_flags $windefs(CONTAINER_INHERIT_ACE)
    }

    if {$opts(recurseobjects)} {
        setbits inherit_flags $windefs(OBJECT_INHERIT_ACE)
    }

    if {$opts(recurseonelevelonly)} {
        setbits inherit_flags $windefs(NO_PROPAGATE_INHERIT_ACE)
    }

    return [list $typecode $inherit_flags $access_mask $sid]
}

#
# Get the ace type (allow, deny etc.)
proc twapi::get_ace_type {ace} {
    return [_ace_type_symbol_to_code [lindex $ace 0]]
}


#
# Set the ace type (allow, deny etc.)
proc twapi::set_ace_type {ace type} {
    return [lreplace $ace 0 0 [_ace_type_code_to_symbol $type]]
}

#
# Get the access rights in an ACE
proc twapi::get_ace_rights {ace {type ""}} {
    return [_access_mask_to_rights [lindex $ace 2] $type]
}

#
# Set the access rights in an ACE
proc twapi::set_ace_rights {ace rights} {
    return [lreplace $ace 2 2 [_access_rights_to_mask $rights]]
}


#
# Get the ACE sid
proc twapi::get_ace_sid {ace} {
    return [lindex $ace 3]
}

#
# Set the ACE sid
proc twapi::set_ace_sid {ace account} {
    return [lreplace $ace 3 3 [map_account_to_sid $account]]
}


#
# Get the inheritance options
proc twapi::get_ace_inheritance {ace} {
    variable windefs
    
    set inherit_opts [list ]
    set inherit_mask [lindex $ace 1]

    lappend inherit_opts -self \
        [expr {($inherit_mask & $windefs(INHERIT_ONLY_ACE)) == 0}]
    lappend inherit_opts -recursecontainers \
        [expr {($inherit_mask & $windefs(CONTAINER_INHERIT_ACE)) != 0}]
    lappend inherit_opts -recurseobjects \
        [expr {($inherit_mask & $windefs(OBJECT_INHERIT_ACE)) != 0}]
    lappend inherit_opts -recurseonlevelonly \
        [expr {($inherit_mask & $windefs(NO_PROPAGATE_INHERIT_ACE)) != 0}]
    lappend inherit_opts -inherited \
        [expr {($inherit_mask & $windefs(INHERITED_ACE)) != 0}]

    return $inherit_opts
}

#
# Set the inheritance options. Unspecified options are not set
proc twapi::set_ace_inheritance {ace args} {
    set unspecified "adfaowtadf"

    array set opts [cmdline::getoptions args \
                        [list \
                             [list self.arg $unspecified] \
                             [list recursecontainers $unspecified] \
                             [list recurseobjects $unspecified] \
                             [list recurseonelevelonly $unspecified]]]
    
    set inherit_flags [lindex $ace 1]
    if {[string compare $opts(self) $unspecified]} {
        if {$opts(self)} {
            resetbits inherit_flags $windefs(INHERIT_ONLY_ACE)
        } else {
            setbits   inherit_flags $windefs(INHERIT_ONLY_ACE)
        }
    }

    foreach {
        opt                 mask
    } {
        recursecontainers   CONTAINER_INHERIT_ACE
        recurseobjects      OBJECT_INHERIT_ACE
        recurseonelevelonly NO_PROPAGATE_INHERIT_ACE
    } {
        if {[string compare $opts($opt) $unspecified]} {
            if {$opts($opt)} {
                setbits inherit_flags $windefs($mask)
            } else {
                resetbits inherit_flags $windefs($mask)
            }
        }
    }

    return [lreplace $ace 1 1 $inherit_flags]
}


#
# Sort ACE's in the standard recommended Win2K order
proc twapi::sort_aces {aces} {
    variable windefs

    _init_ace_type_symbol_to_code_map

    foreach type [array names twapi::_ace_type_symbol_to_code_map] {
        set direct_aces($type) [list ]
        set inherited_aces($type) [list ]
    }
    
    # Sort order is as follows: all direct (non-inherited) ACEs come
    # before all inherited ACEs. Within these groups, the order should be
    # access denied ACEs, access denied ACEs for objects/properties,
    # access allowed ACEs, access allowed ACEs for objects/properties,
    foreach ace $aces {
        set type [get_ace_type $ace]
        if {[lindex $ace 1] & $windefs(INHERITED_ACE)} {
            lappend inherited_aces($type) $ace
        } else {
            lappend direct_aces($type) $ace
        }
    }

    # TBD - check this order
    return [concat \
                $direct_aces(deny) \
                $direct_aces(deny_object) \
                $direct_aces(deny_compound) \
                $direct_aces(deny_callback) \
                $direct_aces(deny_callback_object) \
                $direct_aces(allow) \
                $direct_aces(allow_object) \
                $direct_aces(allow_compound) \
                $direct_aces(allow_callback) \
                $direct_aces(allow_callback_object) \
                $direct_aces(audit) \
                $direct_aces(audit_object) \
                $direct_aces(audit_compound) \
                $direct_aces(audit_callback) \
                $direct_aces(audit_callback_object) \
                $direct_aces(alarm) \
                $direct_aces(alarm_object) \
                $direct_aces(alarm_compound) \
                $direct_aces(alarm_callback) \
                $direct_aces(alarm_callback_object) \
                $inherited_aces(deny) \
                $inherited_aces(deny_object) \
                $inherited_aces(deny_compound) \
                $inherited_aces(deny_callback) \
                $inherited_aces(deny_callback_object) \
                $inherited_aces(allow) \
                $inherited_aces(allow_object) \
                $inherited_aces(allow_compound) \
                $inherited_aces(allow_callback) \
                $inherited_aces(allow_callback_object) \
                $inherited_aces(audit) \
                $inherited_aces(audit_object) \
                $inherited_aces(audit_compound) \
                $inherited_aces(audit_callback) \
                $inherited_aces(audit_callback_object) \
                $inherited_aces(alarm) \
                $inherited_aces(alarm_object) \
                $inherited_aces(alarm_compound) \
                $inherited_aces(alarm_callback) \
                $inherited_aces(alarm_callback_object)]
}



#
# Create a new ACL
proc twapi::new_acl {{aces ""}} {
    variable windefs

    set acl_rev $windefs(ACL_REVISION)
    
    foreach ace $aces {
        set ace_typecode [lindex $ace 0]
        if {[string compare $ace_typecode $windefs(ACCESS_ALLOWED_ACE_TYPE)] &&
            [string compare $ace_typecode $windefs(ACCESS_DENIED_ACE_TYPE)] &&
            [string compare $ace_typecode $windefs(SYSTEM_AUDIT_ACE_TYPE)]} {
            set acl_rev $windefs(ACL_REVISION_DS)
            break
        }
    }

    return [list $acl_rev $aces]
}

#
# Return the list of ACE's in an ACL
proc twapi::get_acl_aces {acl} {
    return [lindex $acl 1]
}

#
# Set the ACE's in an ACL
proc twapi::set_acl_aces {acl aces} {
    # Note, we call new_acl since when ACEs change, the rev may also change
    return [new_acl $aces]
}

#
# Append to the ACE's in an ACL
proc twapi::append_acl_aces {acl aces} {
    return [set_acl_aces $acl [concat [get_acl_aces] $aces]]
}

#
# Prepend to the ACE's in an ACL
proc twapi::prepend_acl_aces {acl aces} {
    return [set_acl_aces $acl [concat $aces [get_acl_aces]]]
}


#
# Arrange the ACE's in an ACL in a standard order
proc twapi::sort_acl_aces {acl} {
    return [set_acl_aces $acl [sort_aces [get_acl_aces $acl]]]
}



#
# Return the ACL revision of an ACL
proc twapi::get_acl_rev {acl} {
    return [lindex $acl 0]
}


#
# Create a new security descriptor
proc twapi::new_security_descriptor {args} {
    set notspecified "adfh98q2ryadf"
    array set opts [cmdline::getoptions args \
                        [list \
                             owner.arg $notspecified \
                             group.arg $notspecified \
                             dacl.arg $notspecified \
                             sacl.arg $notspecified \
                            ]]
    set secd [Twapi_InitializeSecurityDescriptor]

    foreach field {owner group dacl sacl} {
        if {[string compare $opts($field) $notspecified]} {
            set secd [set_security_descriptor_$field $secd $opts($field)]
        }
    }

    return $secd
}


#
# Return the owner in a security descriptor
proc twapi::get_security_descriptor_owner {secd} {
    return [lindex $secd 1]
}


#
# Set the owner in a security descriptor
proc twapi::set_security_descriptor_owner {secd account} {
    set sid [map_account_to_sid $account]
    return [lreplace $secd 1 1 $sid]
}


#
# Return the group in a security descriptor
proc twapi::get_security_descriptor_group {secd} {
    return [lindex $secd 2]
}


#
# Set the group in a security descriptor
proc twapi::set_security_descriptor_group {secd account} {
    set sid [map_account_to_sid $account]
    return [lreplace $secd 2 2 $sid]
}


#
# Return the DACL in a security descriptor
proc twapi::get_security_descriptor_dacl {secd} {
    return [lindex $secd 3]
}


#
# Set the dacl in a security descriptor
proc twapi::set_security_descriptor_dacl {secd acl} {
    return [lreplace $secd 3 3 $acl]
}


#
# Return the SACL in a security descriptor
proc twapi::get_security_descriptor_sacl {secd} {
    return [lindex $secd 4]
}


#
# Set the sacl in a security descriptor
proc twapi::set_security_descriptor_sacl {secd acl} {
    return [lreplace $secd 4 4 $acl]
}


#
# Get the specified security information for the given object
proc twapi::get_resource_security_descriptor {restype name args} {
    variable windefs

    switch -exact -- $restype {
        file      { set type 1 }
        service   { set type 2 }
        printer   { set type 3 }
        registry  { set type 4 }
        share     { set type 5 }
        kernelobj { set type 6 }
        default { error "Unknown resource type '$restype'" }
    }

    array set opts [cmdline::getoptions args {
        owner
        group
        dacl
        sacl
        all
    }]

    set wanted 0

    foreach field {owner group dacl sacl} {
        if {$opts($field) || $opts(all)} {
            set wanted [expr {$wanted | $windefs([string toupper $field]_SECURITY_INFORMATION)}]
        }
    }

    # Note if no options specified, we ask for everything except
    # SACL's which require special privileges
    if {! $wanted} {
        foreach field {owner group dacl} {
            set wanted [expr {$wanted | $windefs([string toupper $field]_SECURITY_INFORMATION)}]
        }
        set opts($field) 1
    }


    set secd [Twapi_GetNamedSecurityInfo $name $type $wanted]

    return $secd
}


#
# Set the specified security information for the given object
proc twapi::set_resource_security_descriptor {type name secd args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        owner
        group
        dacl
        sacl
        all
        protect_dacl
        protect_sacl
    }]

    set mask 0

    if {[min_os_version 5 0]} {
        if {$opts(protect_dacl)} {
            setbits mask $windefs(PROTECTED_DACL_SECURITY_INFORMATION)
        } else {
            setbits mask $windefs(UNPROTECTED_DACL_SECURITY_INFORMATION)
        }

        if {$opts(protect_sacl)} {
            setbits mask $windefs(PROTECTED_SACL_SECURITY_INFORMATION)
        } else {
            setbits mask $windefs(UNPROTECTED_SACL_SECURITY_INFORMATION)
        }
    }

    if {$opts(owner) || $opts(all)} {
        set opts(owner) [get_security_descriptor_owner $secd]
        setbits mask $windefs(OWNER_SECURITY_INFORMATION)
    } else {
        set opts(owner) ""
    }

    if {$opts(group) || $opts(all)} {
        set opts(group) [get_security_descriptor_group $secd]
        setbits mask $windefs(GROUP_SECURITY_INFORMATION)
    } else {
        set opts(group) ""
    }

    if {$opts(dacl) || $opts(all)} {
        set opts(dacl) [get_security_descriptor_dacl $secd]
        setbits mask $windefs(DACL_SECURITY_INFORMATION)
    } else {
        set opts(dacl) null
    }

    if {$opts(sacl) || $opts(all)} {
        set opts(sacl) [get_security_descriptor_sacl $secd]
        setbits mask $windefs(SACL_SECURITY_INFORMATION)
    } else {
        set opts(sacl) null
    }

    SetNamedSecurityInfo $name $type $mask $opts(owner) $opts(group) $opts(dacl) $opts(sacl)
}

#
# Get a token for a user
proc twapi::open_user_token {username password args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        domain.arg
        {type.arg batch}
        {provider.arg default}
    }]

    set typedef "LOGON32_LOGON_[string toupper $opts(type)]"
    if {![info exists windefs($typedef)]} {
        error "Invalid value '$opts(type)' specified for -type option"
    }

    set providerdef "LOGON32_PROVIDER_[string toupper $opts(provider)]"
    if {![info exists windefs($typedef)]} {
        error "Invalid value '$opts(provider)' specified for -provider option"
    }
    
    # If username is of the form user@domain, then domain must not be specified
    # If username is not of the form user@domain, then domain is set to "."
    # if it is empty
    if {[regexp {^([^@]+)@(.+)} $username dummy user domain]} {
        if {[string length $opts(domain)] == 0} {
            error "The -domain option must not be specified when the username is of in UPN format (user@domain)"
        }
    } else {
        if {[string length $opts(domain)] == 0} {
            set opts(domain) "."
        }
    }

    return [LogonUser $username $opts(domain) $password $windefs($typedef) $windefs($providerdef)]
}


#
# Impersonate a user given a token
proc twapi::impersonate_token {token} {
    ImpersonateLoggedOnUser $token
}


#
# Impersonate a user
proc twapi::impersonate_user {args} {
    set token [eval open_user_token $args]
    try {
        impersonate_token $token
    } finally {
        close_token $token
    }
}


#
# Revert to process token
proc twapi::revert_to_self {{opt ""}} {
    RevertToSelf
}


#
# Impersonate self
proc twapi::impersonate_self {level} {
    switch -exact -- $level {
        anonymous      { set level 0 }
        identification { set level 1 }
        impersonation  { set level 2 }
        delegation     { set level 3 }
        default {
            error "Invalid impersonation level $level"
        }
    }
    ImpersonateSelf $level
}


#
# Set a thread token - currently only for current thread
proc twapi::set_thread_token {token} {
    SetThreadToken NULL $token
}

#
# Reset a thread token - currently only for current thread
proc twapi::reset_thread_token {} {
    SetThreadToken NULL NULL
}

#
# Get a new LUID
proc twapi::new_luid {} {
    return [AllocateLocallyUniqueId]
}

#
# TBD - maybe these UUID functions should not be in the security module
# Get a new uuid
proc twapi::new_uuid {{opt ""}} {
    if {[string length $opt]} {
        if {[string equal $opt "-localok"]} {
            set local_ok 1
        } else {
            error "Invalid or unknown argument '$opt'"
        }
    } else {
        set local_ok 0
    }
    return [UuidCreate $local_ok] 
}
proc twapi::nil_uuid {} {
    return [UuidCreateNil]
}



################################################################
# Utility and helper functions



# Returns an sid field from a token
proc twapi::_get_token_sid_field {tok field options} {
    array set opts [cmdline::getoptions options [list name]]
    set owner [GetTokenInformation $tok $twapi::windefs($field)]
    if {$opts(name)} {
        set owner [lookup_account_sid $owner]
    }
    return $owner
}


#
# Map a token attribute mask to list of attribute names
proc twapi::_map_token_attr {attr names prefix} {
    variable windefs

    set attrs [list ]
    foreach attr_name $names {
        set attr_mask $windefs(${prefix}_[string toupper $attr_name])
        if {[expr {$attr & $attr_mask}]} {
            lappend attrs $attr_name
        }
    }

    return $attrs
}

#
# Set/reset the given bits in the usri3_flags field for a user account
# mask indicates the mask of bits to set. values indicates the values
# of those bits
proc twapi::_change_usri3_flags {username mask values args} {
    eval set [cmdline::getoptions args {
        system.arg
    }]

    # Get current flags
    array set data [NetUserGetInfo $opts(system) $account 1]

    # Turn off mask bits and write flags back
    set flags [expr {$data(usri3_flags) & (~ $mask)}]
    # Set the specified bits
    set flags [expr {$flags | ($values & $mask)}]

    # Write new flags back
    Twapi_NetUserSetInfo_flags $system $username $flags
}


#
# Map a set of access right symbols to a flag. Concatenates
# all the arguments, and then OR's the individual elements. Each
# element may either be a integer or one of the access rights
proc twapi::_access_rights_to_mask {args} {
    variable windefs

    set rights 0
    foreach right [concat $args] {
        if {![string is integer $right]} {
            if {[catch {set right $windefs([string toupper $right])}]} {
                error "Invalid access right symbol '$right'"
            }
        }
        set rights [expr {$rights | $right}]
    }

    return $rights
}


#
# Map an access mask to a set of rights
proc twapi::_access_mask_to_rights {access_mask {type ""}} {
    variable windefs

    set rights [list ]

    # The returned list will include rights that map to multiple bits
    # as well as the individual bits. We first add the multiple bits
    # and then the individual bits (since we clear individual bits
    # after adding)

    #
    # Check standard multiple bit masks
    #
    foreach x {STANDARD_RIGHTS_REQUIRED STANDARD_RIGHTS_READ STANDARD_RIGHTS_WRITE STANDARD_RIGHTS_EXECUTE STANDARD_RIGHTS_ALL SPECIFIC_RIGHTS_ALL} {
        if {($windefs($x) & $access_mask) == $windefs($x)} {
            lappend rights [string tolower $x]
        }
    }

    #
    # Check type specific multiple bit masks
    #
    switch -exact -- $type {
        file {
            set masks [list FILE_ALL_ACCESS FILE_GENERIC_READ FILE_GENERIC_WRITE FILE_GENERIC_EXECUTE]
        }
        pipe {
            set masks [list FILE_ALL_ACCESS]
        }
        service {
            set masks [list SERVICE_ALL_ACCESS]
        }
        registry {
            set masks [list KEY_READ KEY_WRITE KEY_EXECUTE KEY_ALL_ACCESS]
        }
        process {
            set masks [list PROCESS_ALL_ACCESS]
        }
        thread {
            set masks [list THREAD_ALL_ACCESS]
        }
        token {
            set masks [list TOKEN_READ TOKEN_WRITE TOKEN_EXECUTE]
            # TOKEN_ALL_ACCESS depends on platform
            if {[min_os_version 5 0]} {
                set token_all_access $windefs(TOKEN_ALL_ACCESS_WIN2K)
            } else {
                set token_all_access $windefs(TOKEN_ALL_ACCESS_WIN2K)
            }
            if {($token_all_access & $access_mask) == $token_all_access} {
                lappend rights "token_all_access"
            }
        }
        default {
            set masks [list ]
        }
    }

    foreach x $masks {
        if {($windefs($x) & $access_mask) == $windefs($x)} {
            lappend rights [string tolower $x]
        }
    }


    #
    # OK, now map individual bits

    # First map the common bits
    foreach x {DELETE READ_CONTROL WRITE_DAC WRITE_OWNER SYNCHRONIZE} {
        if {$windefs($x) & $access_mask} {
            lappend rights [string tolower $x]
            set access_mask [expr {$access_mask & ~ $windefs($x)}]
        }
    }

    # Then the generic bits
    foreach x {GENERIC_READ GENERIC_WRITE GENERIC_EXECUTE GENERIC_ALL} {
        if {$windefs($x) & $access_mask} {
            lappend rights [string tolower $x]
            set access_mask [expr {$access_mask & ~ $windefs($x)}]
        }
    }

    # Then the type specific
    switch -exact -- $type {
        file {
            set masks {
                FILE_READ_DATA
                FILE_WRITE_DATA
                FILE_APPEND_DATA
                FILE_READ_EA
                FILE_WRITE_EA
                FILE_EXECUTE
                FILE_DELETE_CHILD
                FILE_READ_ATTRIBUTES
                FILE_WRITE_ATTRIBUTES
            }
        }
        pipe {
            set masks {
                FILE_READ_DATA
                FILE_WRITE_DATA
                FILE_CREATE_PIPE_INSTANCE
                FILE_READ_ATTRIBUTES
                FILE_WRITE_ATTRIBUTES
            }
        }
        service {
            set masks {
                SERVICE_QUERY_CONFIG
                SERVICE_CHANGE_CONFIG
                SERVICE_QUERY_STATUS
                SERVICE_ENUMERATE_DEPENDENTS
                SERVICE_START
                SERVICE_STOP
                SERVICE_PAUSE_CONTINUE
                SERVICE_INTERROGATE
                SERVICE_USER_DEFINED_CONTROL
            }
        }
        registry {
            set masks {
                KEY_QUERY_VALUE
                KEY_SET_VALUE
                KEY_CREATE_SUB_KEY
                KEY_ENUMERATE_SUB_KEYS
                KEY_NOTIFY
                KEY_CREATE_LINK
                KEY_WOW64_32KEY
                KEY_WOW64_64KEY
                KEY_WOW64_RES
            }
        }
        process {
            set masks {
                PROCESS_TERMINATE
                PROCESS_CREATE_THREAD
                PROCESS_SET_SESSIONID
                PROCESS_VM_OPERATION
                PROCESS_VM_READ
                PROCESS_VM_WRITE
                PROCESS_DUP_HANDLE
                PROCESS_CREATE_PROCESS
                PROCESS_SET_QUOTA
                PROCESS_SET_INFORMATION
                PROCESS_QUERY_INFORMATION
                PROCESS_SUSPEND_RESUME
            }
        }
        thread {
            set masks {
                THREAD_TERMINATE
                THREAD_SUSPEND_RESUME
                THREAD_GET_CONTEXT
                THREAD_SET_CONTEXT
                THREAD_SET_INFORMATION
                THREAD_QUERY_INFORMATION
                THREAD_SET_THREAD_TOKEN
                THREAD_IMPERSONATE
                THREAD_DIRECT_IMPERSONATION
            }
        }
        token {
            set masks {
                TOKEN_ASSIGN_PRIMARY
                TOKEN_DUPLICATE
                TOKEN_IMPERSONATE
                TOKEN_QUERY
                TOKEN_QUERY_SOURCE
                TOKEN_ADJUST_PRIVILEGES
                TOKEN_ADJUST_GROUPS
                TOKEN_ADJUST_DEFAULT
                TOKEN_ADJUST_SESSIONID
            }
        }
        default {
            set masks [list ]
        }
    }

    foreach x $masks {
        if {$windefs($x) & $access_mask} {
            lappend rights [string tolower $x]
            set access_mask [expr {$access_mask & ~ $windefs($x)}]
        }
    }

    # Finally add left over bits if any
    for {set i 0} {$i < 32} {incr i} {
        set x [expr {1 << $i}]
        if {$access_mask & $x} {
            lappend rights [format 0x%.8X $x]
        }
    }

    return $rights
}


#
# Map an ace type symbol (eg. allow) to the underlying ACE type code
proc twapi::_ace_type_symbol_to_code {type} {
    _init_ace_type_symbol_to_code_map
    return $::twapi::_ace_type_symbol_to_code_map($type)
}


#
# Map an ace type code to an ACE type symbol
proc twapi::_ace_type_code_to_symbol {type} {
    _init_ace_type_symbol_to_code_map
    return $::twapi::_ace_type_code_to_symbol_map($type)
}


# Init the arrays used for mapping ACE type symbols to codes and back
proc twapi::_init_ace_type_symbol_to_code_map {} {
    if {[info exists ::twapi::_ace_type_symbol_to_code_map]} {
        return
    }

    # Define the array. Be careful to "normalize" the integer values
    array set ::twapi::_ace_type_symbol_to_code_map \
        [list \
             allow [expr { $windefs(ACCESS_ALLOWED_ACE_TYPE) + 0 }] \
             deny [expr  { $windefs(ACCESS_DENIED_ACE_TYPE) + 0 }] \
             audit [expr { $windefs(SYSTEM_AUDIT_ACE_TYPE) + 0 }] \
             alarm [expr { $windefs(SYSTEM_ALARM_ACE_TYPE) + 0 }] \
             allow_compound [expr { $windefs(ACCESS_ALLOWED_COMPOUND_ACE_TYPE) + 0 }] \
             allow_object [expr   { $windefs(ACCESS_ALLOWED_OBJECT_ACE_TYPE) + 0 }] \
             deny_object [expr    { $windefs(ACCESS_DENIED_OBJECT_ACE_TYPE) + 0 }] \
             audit_object [expr   { $windefs(SYSTEM_AUDIT_OBJECT_ACE_TYPE) + 0 }] \
             alarm_object [expr   { $windefs(SYSTEM_ALARM_OBJECT_ACE_TYPE) + 0 }] \
             allow_callback [expr { $windefs(ACCESS_ALLOWED_CALLBACK_ACE_TYPE) + 0 }] \
             deny_callback [expr  { $windefs(ACCESS_DENIED_CALLBACK_ACE_TYPE) + 0 }] \
             allow_callback_object [expr { $windefs(ACCESS_ALLOWED_CALLBACK_OBJECT_ACE_TYPE) + 0 }] \
             deny_callback_object [expr  { $windefs(ACCESS_DENIED_CALLBACK_OBJECT_ACE_TYPE) + 0 }] \
             audit_callback [expr { $windefs(SYSTEM_AUDIT_CALLBACK_ACE_TYPE) + 0 }] \
             alarm_callback [expr { $windefs(SYSTEM_ALARM_CALLBACK_ACE_TYPE) + 0 }] \
             audit_callback_object [expr { $windefs(SYSTEM_AUDIT_CALLBACK_OBJECT_ACE_TYPE) + 0 }] \
             alarm_callback_object [expr { $windefs(SYSTEM_ALARM_CALLBACK_OBJECT_ACE_TYPE) + 0 }] \
                 ]

    # Now define the array in the other direction
    foreach {sym code} [array get ::twapi::_ace_type_symbol_to_code_map] {
        set ::twapi::_ace_type_code_to_symbol_map($code) $sym
    }
}
