#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

package require TclExcept
package require cmdline
package require twapi

namespace eval twapi {
} 

# Locks the scm database and returns a handle to the lock
proc twapi::lock_scm_db {args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]

    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(SC_MANAGER_LOCK)]
    try {
        set lock [LockServiceDatabase $scm]
    } finally {
        CloseServiceHandle $scm
    }
    return $lock
}

# Unlocks the scm database
proc twapi::unlock_scm_db {lock} {
    UnlockServiceDatabase $lock
}

# Query the lock status of the scm database. Returns 1/0 depending on whether
# database is locked. $v_lockinfo will contain a two element list of
# {lockowner time_locked}
proc twapi::query_scm_db_lock_status {v_lockinfo args} {
    variable windefs

    upvar $v_lockinfo lockinfo

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(SC_MANAGER_QUERY_LOCK_STATUS)]
    try {
        array set lock_status [QueryServiceLockStatus $scm]
        set lockinfo [list $lock_status(lpLockOwner) $lock_status(dwLockDuration)]
    } finally {
        CloseServiceHandle $scm
    }

    return $lock_status(fIsLocked)
}


#
# Return 1/0 depending on whether the given service exists
# $name may be either the internal or display name
proc twapi::service_exists {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(STANDARD_RIGHTS_READ)]
 
    try {
        GetServiceKeyName $scm $name
        set exists 1
    } catch msg {
        default {
            if {! [_no_such_service_error $::errorCode]} {
                throw
            }
            try {
                GetServiceDisplayName $scm $name
                set exists 1
            } catch msg {
                default {
                    if {! [_no_such_service_error $::errorCode]} {
                        throw
                    }
                    set exists 0
                }
            } 
        }
    } finally {
        CloseServiceHandle $scm
    }

    return $exists
}

#
# Utility proc to indicate if error_code corresponds to a Windows
# no such service error.
proc twapi::_no_such_service_error {error_code} {
    if {[string equal [lindex $error_code 0] "WINDOWS"]
        && ([lindex $error_code 1] == 1060)} {
        return 1
    } else {
        return 0
    }
}

#
# Create a service of the specified name
proc twapi::create_service {name command args} {
    variable windefs

    array set opts [cmdline::getoptions args \
                        [list \
                             [list displayname.arg    $name] \
                             [list servicetype.arg     win32_own_process] \
                             [list interactive.arg     0] \
                             [list starttype.arg       auto_start] \
                             [list errorcontrol.arg    normal] \
                             loadordergroup.arg \
                             dependencies.arg \
                             account.arg \
                             password.arg \
                             system.arg \
                             database.arg]]
    

    if {[string length $command] == 0} {
        error "The executable path must not be null when creating a service"
    }
    set opts(command) $command

    # Convert options to system parameter values
    _parse_service_config_args opts winparams
    
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(SC_MANAGER_CREATE_SERVICE)]
    try {
        set svch [CreateService \
                      $scm \
                      $name \
                      $winparams(displayname) \
                      $windefs(SERVICE_ALL_ACCESS) \
                      $winparams(servicetype) \
                      $winparams(starttype) \
                      $winparams(errorcontrol) \
                      $winparams(command) \
                      $winparams(loadordergroup) \
                      NULL \
                      $winparams(dependencies) \
                      $winparams(account) \
                      $winparams(password)]
        
        CloseServiceHandle $svch
        
    } finally {
        CloseServiceHandle $scm
    }
    
    return
}


# Delete the given service
proc twapi::delete_service {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set opts(scm_priv) DELETE
    set opts(svc_priv) DELETE
    set opts(proc)     twapi::DeleteService
    
    _service_fn_wrapper $name opts

    return
}


# Get the internal name of a service
proc twapi::get_service_internal_name {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(STANDARD_RIGHTS_READ)]


    try {
        if {[catch {GetServiceKeyName $scm $name} internal_name]} {
            # Maybe this is an internal name itself
            GetServiceDisplayName $scm $name; # Will throw an error if not internal name
            set internal_name $name
        }
    } finally {
        CloseServiceHandle $scm
    }

    return $internal_name
}

# Get the display name of a service
proc twapi::get_service_display_name {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(STANDARD_RIGHTS_READ)]


    try {
        if {[catch {GetServiceDisplayName $scm $name} display_name]} {
            # Maybe this is an display name itself
            GetServiceKeyName $scm $name; # Will throw an error if not display name
            set display_name $name
        }
    } finally {
        CloseServiceHandle $scm
    }

    return $display_name
}


# Start the specified service
proc twapi::start_service {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        system.arg
        database.arg
        params.arg
        {wait.arg 0}
    }]
    set opts(svc_priv) SERVICE_START
    set opts(proc)     twapi::StartService
    set opts(args)     [list $opts(params)]
    unset opts(params)

    try {
        _service_fn_wrapper $name opts
    } catch msg {
        default {
            # Error 1056 means service already running
            if {[string compare [lindex $::errorCode 0] WINDOWS] ||
                ([lindex $::errorCode 1] != 1056)} {
                throw
            }
        }
    }
    return [wait {twapi::get_service_state $name -system $opts(system) -database $opts(database)} running $opts(wait)]
}


# Send a control code to the specified service
proc twapi::control_service {name code access finalstate args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        system.arg
        database.arg
        ignorecodes.arg
        {wait.arg 0}
    }]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(STANDARD_RIGHTS_READ)]
    
    try {
        set svch [OpenService $scm $name $access]
    } finally {
        CloseServiceHandle $scm
    }

    SERVICE_STATUS svc_status
    try {
        ControlService $svch $code svc_status
    } catch msg {
        default {
            if {[string compare [lindex $::errorCode 0] WINDOWS] ||
                ([lsearch -exact -integer $opts(ignorecodes) [lindex $::errorCode 1]] < 0)} {
                throw
            }
        }
    } finally {
        svc_status -delete
        CloseServiceHandle $svch
    }

    if {[string length $finalstate]} {
        # Wait until service is in specified state
        return [wait {twapi::get_service_state $name -system $opts(system) -database $opts(database)} $finalstate $opts(wait)]
    } else {
        return 0
    }
}

proc twapi::stop_service {name args} {
    variable windefs
    eval [list control_service $name \
              $windefs(SERVICE_CONTROL_STOP) $windefs(SERVICE_STOP) stopped -ignorecodes 1062] $args
}

proc twapi::pause_service {name args} {
    variable windefs
    eval [list control_service $name \
              $windefs(SERVICE_CONTROL_PAUSE) \
              $windefs(SERVICE_PAUSE_CONTINUE) paused] $args
}

proc twapi::continue_service {name args} {
    variable windefs
    eval [list control_service $name \
              $windefs(SERVICE_CONTROL_CONTINUE) \
              $windefs(SERVICE_PAUSE_CONTINUE) running] $args
}

proc twapi::interrogate_service {name args} {
    variable windefs
    eval [list control_service $name \
              $windefs(SERVICE_CONTROL_INTERROGATE) \
              $windefs(SERVICE_INTERROGATE) ""] $args
    return
}


# Retrieve status information for a service
proc twapi::get_service_status {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {system.arg database.arg}]
    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(STANDARD_RIGHTS_READ)]

    try {
        set svch [OpenService $scm $name $windefs(SERVICE_QUERY_STATUS)]
    } finally {
        # Do not need SCM anymore
        CloseServiceHandle $scm
    }

    SERVICE_STATUS svc_status
    try {
        QueryServiceStatus $svch svc_status
    } catch msg {
        default {
            svc_status -delete
            throw
        }
    } finally {
        CloseServiceHandle $svch
    }

    return [_format_SERVICE_STATUS svc_status -delete]
}


# Get the state of the service
proc twapi::get_service_state {name args} {
    return [kl_get [eval [list get_service_status $name] $args] state]
}


# Get the current configuration for a service
proc twapi::get_service_configuration {name args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        system.arg
        database.arg
        all
        servicetype
        interactive
        errorcontrol
        starttype
        command
        loadordergroup
        account
        displayname
        dependencies
    }]

    set opts(svc_priv) SERVICE_QUERY_CONFIG
    set opts(proc)     twapi::QueryServiceConfig
    array set svc_config [_service_fn_wrapper $name opts]

    # Get the service type info
    foreach {servicetype interactive} \
        [_map_servicetype_code $svc_config(dwServiceType)] break

    set result [list ]
    if {$opts(all) || $opts(servicetype)} {
        lappend result -servicetype $servicetype
    }
    if {$opts(all) || $opts(interactive)} {
        lappend result -interactive $interactive
    }
    if {$opts(all) || $opts(errorcontrol)} {
        lappend result -errorcontrol [_map_errorcontrol_code $svc_config(dwErrorControl)]
    }
    if {$opts(all) || $opts(starttype)} {
        lappend result -starttype [_map_starttype_code $svc_config(dwStartType)]
    }
    if {$opts(all) || $opts(command)} {
        lappend result -command $svc_config(lpBinaryPathName)
    }
    if {$opts(all) || $opts(loadordergroup)} {
        lappend result -loadordergroup $svc_config(lpLoadOrderGroup)
    }
    if {$opts(all) || $opts(account)} {
        lappend result -account $svc_config(lpServiceStartName)
    }
    if {$opts(all) || $opts(displayname)} {
        lappend result -displayname    $svc_config(lpDisplayName)
    }
    if {$opts(all) || $opts(dependencies)} {
        lappend result -dependencies $svc_config(lpDependencies)
    }
    # Note currently we do not return svc_config(dwTagId)

    return $result
}

# Sets a service configuration
proc twapi::set_service_configuration {name args} {
    variable windefs

    # Remember which arguments were actually specified
    array set specified_args $args

    # Make sure -servicetype specified if -interactive is specified.
    # (This is an arbitrary restriction based on some misfeatures in the code
    # parsing code)
    if {[info exists specified_args(-interactive)]
        && ![info exists specified_args(-servicetype)]} {
        error "Option -servicetype must also be specified if -interactive option is present."
    }

    # Now parse arguments, filling in defaults
    array set opts [cmdline::getoptions args \
                        [list \
                             [list displayname.arg  $name] \
                             [list servicetype.arg   win32_own_process] \
                             [list interactive.arg   0] \
                             [list starttype.arg     auto_start] \
                             [list errorcontrol.arg  normal] \
                             command.arg         \
                             loadordergroup.arg \
                             dependencies.arg \
                             account.arg \
                             password.arg \
                             system.arg \
                             database.arg]]


    # Convert options to system parameter values
    _parse_service_config_args opts winparams

    # Now replace any options that were not specified with "no change"
    # tokens. Note -interactive is not listed here, since it is included
    # in servicetype at this point
    foreach param {servicetype starttype errorcontrol} {
        if {![info exists specified_args(-$param)]} {
            set winparams($param) $windefs(SERVICE_NO_CHANGE)
        }
    }
    foreach param {command loadordergroup dependencies account password displayname} {
        if {![info exists specified_args(-$param)]} {
            set winparams($param) $twapi::nullptr
        }
    }

    set opts(scm_priv) STANDARD_RIGHTS_READ
    set opts(svc_priv) SERVICE_CHANGE_CONFIG
    set opts(proc)     twapi::ChangeServiceConfig
    set opts(args) \
        [list \
             $winparams(servicetype) \
             $winparams(starttype) \
             $winparams(errorcontrol) \
             $winparams(command) \
             $winparams(loadordergroup) \
             NULL \
             $winparams(dependencies) \
             $winparams(account) \
             $winparams(password) \
             $winparams(displayname)]

    _service_fn_wrapper $name opts

    return
}


# Get status for the specified service types
proc twapi::get_multiple_service_status {args} {
    variable windefs

    set service_types [list \
                           kernel_driver \
                           file_system_driver \
                           adapter \
                           recognizer_driver \
                           win32_own_process \
                           win32_share_process]
    set switches [concat $service_types \
                      [list active inactive] \
                      [list system.arg database.arg]]
    array set opts [cmdline::getoptions args $switches]

    set servicetype 0
    foreach type $service_types {
        if {$opts($type)} {
            set servicetype [expr { $servicetype
                                    | $windefs(SERVICE_[string toupper $type])
                                }]
        }
    }
    if {$servicetype == 0} {
        # No type specified, return all
        set servicetype [expr {$windefs(SERVICE_KERNEL_DRIVER)
                               | $windefs(SERVICE_FILE_SYSTEM_DRIVER)
                               | $windefs(SERVICE_ADAPTER)
                               | $windefs(SERVICE_RECOGNIZER_DRIVER)
                               | $windefs(SERVICE_WIN32_OWN_PROCESS)
                               | $windefs(SERVICE_WIN32_SHARE_PROCESS)}]
    }

    set servicestate 0
    if {$opts(active)} {
        set servicestate [expr {$servicestate |
                                $windefs(SERVICE_ACTIVE)}]
    }
    if {$opts(inactive)} {
        set servicestate [expr {$servicestate |
                                $windefs(SERVICE_INACTIVE)}]
    }
    if {$servicestate == 0} {
        # No state specified, include all
        set servicestate $windefs(SERVICE_STATE_ALL)
    }
    
    set servicelist [list ]

    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs(SC_MANAGER_ENUMERATE_SERVICE)]
    try {
        set status_recs [EnumServicesStatus $scm $servicetype $servicestate]
    } finally {
        CloseServiceHandle $scm
    }

    foreach status_rec $status_recs {
        lappend servicelist [_format_status_record $status_rec]
    }

    return $servicelist
}


# Get status for the dependents of the specified service
proc twapi::get_dependent_service_status {name args} {
    variable windefs

    array set opts [cmdline::getoptions args \
                        [list active inactive system.arg database.arg]]

    set servicestate 0
    if {$opts(active)} {
        set servicestate [expr {$servicestate |
                                $windefs(SERVICE_ACTIVE)}]
    }
    if {$opts(inactive)} {
        set servicestate [expr {$servicestate |
                                $windefs(SERVICE_INACTIVE)}]
    }
    if {$servicestate == 0} {
        # No state specified, include all
        set servicestate $windefs(SERVICE_STATE_ALL)
    }
    
    set opts(svc_priv) SERVICE_ENUMERATE_DEPENDENTS
    set opts(proc)     twapi::EnumDependentServices
    set opts(args)     [list $servicestate]
    set status_recs [_service_fn_wrapper $name opts]

    set servicelist [list ]
    foreach status_rec $status_recs {
        lappend servicelist [_format_status_record $status_rec]
    }
        
    return $servicelist
}


# Map an integer service type code into a list consisting of 
# {SERVICETYPESYMBOL BOOLEAN}. If there is not symbolic service type
# for the service, just the integer code is returned. The BOOLEAN
# is 1/0 depending on whether the service type code is interactive
proc twapi::_map_servicetype_code {servicetype} {
    variable windefs

    set interactive [expr {($servicetype & $windefs(SERVICE_INTERACTIVE_PROCESS)) != 0}]
    set servicetype [expr {$servicetype & (~$windefs(SERVICE_INTERACTIVE_PROCESS))}]

    # Map service type integer code to symbol, *if possible*
    set service_syms {
        win32_own_process win32_share_process kernel_driver
        file_system_driver adapter recognizer_driver
    }
    set servicetype [code_to_symbol $servicetype $service_syms]
    
    return [list $servicetype $interactive]
}

# Map a start type code into a symbol. Returns the integer code if
# no mapping possible
proc twapi::_map_starttype_code {code} {
    return [code_to_symbol \
                $code {auto_start boot_start demand_start disabled system_start}]
}


# Map a error control code into a symbol. Returns the integer code if
# no mapping possible
proc twapi::_map_errorcontrol_code {code} {
    return [code_to_symbol \
                $code {ignore normal severe critical} "SERVICE_ERROR_"]
}


# Map a service state code to a symbol 
proc twapi::_map_state_code {code} {
    # Map state integer code to symbol *if possible*
    set states {
        stopped start_pending stop_pending running continue_pending
        pause_pending paused
    }
    set state [code_to_symbol $code $states]
}


#
# Format a service status object returned from SWIG
proc twapi::_format_SERVICE_STATUS {ss_obj args} {
    variable windefs

    array set opts [cmdline::getoptions args {delete}]

    set status_rec \
        [list \
             dwServiceType [$ss_obj cget -dwServiceType] \
             dwCurrentState [$ss_obj cget -dwCurrentState] \
             dwControlsAccepted [$ss_obj cget -dwControlsAccepted] \
             dwWin32ExitCode    [$ss_obj cget -dwWin32ExitCode] \
             dwServiceSpecificExitCode [$ss_obj cget -dwServiceSpecificExitCode] \
             dwCheckPoint        [$ss_obj cget -dwCheckPoint] \
             dwWaitHint         [$ss_obj cget -dwWaitHint]]

    set status_list [_format_status_record $status_rec]

    if {$opts(delete)} {
        $ss_obj -delete
    }

    return $status_list
}

#
# Format a service status list record 
# {dwServiceType 1 dwCurrentState 2....}
# gets formatted as 
# {-servicetype win32_own_process -state running ...}
proc twapi::_format_status_record {status_rec} {
    variable windefs

    array set ss $status_rec

    # Get the service type and strip off the interactive flag
    foreach {servicetype interactive} \
        [_map_servicetype_code $ss(dwServiceType)] break

    # Map state integer code to symbol *if possible*
    set state [_map_state_code $ss(dwCurrentState)]

    set status_rec [list ]
    if {[info exists ss(lpServiceName)]} {
        lappend status_rec name $ss(lpServiceName)
    }
    if {[info exists ss(lpDisplayName)]} {
        lappend status_rec displayname $ss(lpDisplayName)
    }
    lappend status_rec \
        servicetype  $servicetype \
        interactive  $interactive \
        state        $state \
        controls_accepted $ss(dwControlsAccepted) \
        exitcode     $ss(dwWin32ExitCode) \
        service_code $ss(dwServiceSpecificExitCode) \
        checkpoint   $ss(dwCheckPoint) \
        wait_hint    $ss(dwWaitHint)
    return $status_rec
}


#
# Parse the configuration parameters for a service and convert them
# to an internal form suitable for passing to the system function
# inv refers to an array containing the input arguments. outv is
# an array that should hold the result
proc twapi::_parse_service_config_args {inv outv} {
    variable windefs
    upvar $inv  in
    upvar $outv out

    # NOTE: Valid values in the lists below are based on the symbols in the
    # windefs table

    set valid_servicetypes [list \
                                win32_own_process \
                                win32_share_process \
                                file_system_driver \
                                kernel_driver]
    
    set valid_starttypes [list \
                              auto_start \
                              boot_start \
                              demand_start \
                              disabled \
                              system]
    
    set valid_errorcontrols [list \
                                 ignore \
                                 normal \
                                 severe \
                                 critical]
    
    # Validate arguments
    
    # As an aside, -interactive is not marked as a boolean in the arg list
    # above so that values returned from various status function can be
    # treated uniformly by applications when modifying and passing back
    # to configuration functions
    if {[catch {expr { $in(interactive) ? 1 : 0 } } out(interactive)]} {
        error "Invalid -interactive value. Should be a boolean value"
    }
    
    if {[lsearch -exact $valid_servicetypes $in(servicetype)] < 0} {
        error "Invalid -servicetype value $in(servicetype). Should be one of [join $valid_servicetypes ,]."
    }
    
    if {[lsearch -exact $valid_errorcontrols $in(errorcontrol)] < 0} {
        error "Invalid -errorcontrol value $in(errorcontrol). Should be one of [join $valid_errorcontrols ,]."
    }
    
    if {[lsearch -exact $valid_starttypes $in(starttype)] < 0} {
        error "Invalid -starttype value $in(starttype). Should be one of [join $valid_starttypes ,]."
    }
    
    switch -exact -- $in(servicetype) {
        file_system_driver -
        kernel_driver {
            if {$in(interactive)} {
                error "Option -interactive cannot be specified when -servicetype is $in(servicetype)."
            }
        }
        default {
            if {$in(interactive) && [string length $in(account)]} {
                error "Option -interactive cannot be specified with the -account option as interactive services must run under the LocalSystem account."
            }
            if {[string equal $in(starttype) "boot_start"]
                || [string equal $in(starttype) "system_start"]} {
                error "Option -starttype value must be one of auto_start, demand_start or disabled when -servicetype is '$in(servicetype)'."
            }
        }
    }
    
    # Map keywords to integer values
    set out(servicetype)  $windefs(SERVICE_[string toupper $in(servicetype)])
    set out(starttype)    $windefs(SERVICE_[string toupper $in(starttype)])
    set out(errorcontrol) $windefs(SERVICE_ERROR_[string toupper $in(errorcontrol)])

    # If interactive, add the flag to the service type
    if {$out(interactive)} {
        set out(servicetype) \
            [expr {$out(servicetype) | $windefs(SERVICE_INTERACTIVE_PROCESS)}]
    }

    foreach key {
        displayname command loadordergroup dependencies account password
    } {
        set out($key) $in($key)
    }

    # Ignore password if username not specified
    # TBD - IS this desired when called by set_service_configuration?
    if {[string length $out(account)] == 0} {
        set out(password) ""
    } else {
        # If domain/system not specified, tack on ".\" for local system
        if {[string first \\ $out(account)] < 0} {
            set out(account) ".\\$out(account)"
        }
    }
    
    return
}

    

# Maps an integer code to a symbolic name if possible, else returns the
# integer code
proc twapi::code_to_symbol {code symlist {prefix "SERVICE_"}} {
    variable windefs
    foreach sym $symlist {
        if {$code == $windefs(${prefix}[string toupper $sym])} {
            return $sym
        }
    }
    return $code
}


#
# Standard template for calling a service function. v_opts should refer
# to an array with the following elements:
# opts(system) - target system. Must be specified
# opts(database) - target database. Must be specified
# opts(scm_priv) - requested privilege when opening SCM. STANDARD_RIGHTS_READ
#   is used if unspecified
# opts(svc_priv) - requested privilege when opening service. Must be present
# opts(proc) - proc/function to call. The first arg is the service handle
# opts(args) - additional arguments to pass to the function.
#   Empty if unspecified
# 
proc twapi::_service_fn_wrapper {name v_opts} {
    variable windefs
    upvar $v_opts opts

    # Use STANDARD_RIGHTS_READ for SCM if not specified
    set scm_priv STANDARD_RIGHTS_READ
    catch {set scm_priv $opts(scm_priv)}

    set scm [OpenSCManager $opts(system) $opts(database) \
                 $windefs($scm_priv)]
    try {
        set svch [OpenService $scm $name $windefs($opts(svc_priv))]
    } finally {
        # No need for scm handle anymore
        CloseServiceHandle $scm
    }

    set proc_args [list ]
    catch {set proc_args $opts(args)}
    try {
        set results [eval [list $opts(proc) $svch] $proc_args]
    } finally {
        CloseServiceHandle $svch
    }
    
    return $results
}


