#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

package require TclExcept
package require cmdline
package require twapi

namespace eval twapi {
}


#
# Create a network share
proc twapi::new_share {sharename path args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        system.arg
        {type.arg "file"}
        comment.arg
        {max_conn -1}
        {secd.arg ""}
    }]

    NetShareAdd $opts(system) \
        $sharename \
        [_share_type_symbols_to_code $opts(type)] \
        $opts(comment) \
        $opts(max_conn) \
        $path $opts(secd)
}

#
# Delete a network share
proc twapi::delete_share {sharename args} {
    array set opts [cmdline::getoptions args {system.arg}]
    NetShareDel $opts(system) $net_name 0
}

#
# Enumerate network shares
proc twapi::get_shares {args} {
    variable windefs

    array set opts [cmdline::getoptions args {
        system.arg
        type.arg
        excludespecial
    }]

    if {$opts(type) != ""} {
        set type_filter [_share_type_symbols_to_code $opts(type)]
    } else {
        set type_filter ""
    }

    set shares [list ]
    foreach share [Twapi_NetShareEnum $opts(system)] {
        foreach {name type comment} $share break
        set special [expr {$type & ($windefs(STYPE_SPECIAL) | $windefs(STYPE_TEMPORARY))}]
        if {$special && $opts(excludespecial)} {
            continue
        }
        # We need the special cast to int because else operands get promoted
        # to 64 bits as the hex is treated as an unsigned value
        set type [expr {int($type & ~ $special)}]
        if {([string length $type_filter] == 0) || ($type == $type_filter)} {
            lappend shares $name
        }
    }

    return $shares
}


#
# Get details about a share
proc twapi::get_share_info {sharename args} {
    array set opts [cmdline::getoptions args {
        system.arg
        all
        name
        type
        path
        comment
        max_conn
        current_conn
        secd
    }]

    if {$opts(all)} {
        foreach opt {name type path comment max_conn current_conn secd} {
            set opts($opt) 1
        }
    }

    set level 0

    if {$opts(name) || $opts(type) || $opts(comment)} {
        set level 1
    }

    if {$opts(max_conn) || $opts(current_conn) || $opts(path)} {
        set level 2
    }

    if {$opts(secd)} {
        set level 502
    }

    if {! $level} {
        return
    }

    set shareinfo [NetShareGetInfo $opts(system) $sharename $level]

    set result [list ]
    if {$opts(name)} {
        lappend result -name [lindex $shareinfo 0]
    }
    if {$opts(type)} {
        lappend result -type [_share_type_code_to_symbols [lindex $shareinfo 1]]
    }
    if {$opts(comment)} {
        lappend result -comment [lindex $shareinfo 2]
    }
    if {$opts(max_conn)} {
        lappend result -max_conn [lindex $shareinfo 4]
    }
    if {$opts(current_conn)} {
        lappend result -current_conn [lindex $shareinfo 5]
    }
    if {$opts(path)} {
        lappend result -path [lindex $shareinfo 6]
    }
    if {$opts(secd)} {
        lappend result -secd [lindex $shareinfo 9]
    }
    
    return $result
}


#
# Set a share configuration
proc twapi::set_share_info {sharename args} {
    set unspecified "adfafgadg"
    array set opts [cmdline::getoptions args \
                        [list \
                             system.arg \
                             [list comment.arg $unspecified] \
                             [list max_conn.arg $unspecified] \
                             [list secd.arg $unspecified] \
                            ]]

    # First get the current config so we can change specified fields
    # and write back
    array set shareinfo [get_share_info $sharename -system $opts(system) \
                             -comment -max_conn -secd]
    foreach field {comment max_conn secd} {
        if {[string compare $opts($field) $unspecified]} {
            set shareinfo(-$field) $opts($field)
        }
    }

    NetShareSetInfo $opts(system) $sharename $shareinfo(comment) \
        $shareinfo(max_conn) $shareinfo(secd)
}


#
# Get list of remote shares
proc twapi::get_connected_shares {} {
    return [NetUseEnum]
}


#
# Connect to a share
proc twapi::connect_share {remoteshare args} {
    array set opts [cmdline::getoptions args {
        {type.arg  "disk"} 
        localdevice.arg
        provider.arg
        password.arg
        nopassword
        defaultpassword
        user.arg
        {window.arg 0}
        interactive
        prompt
        updateprofile
        commandline
    }]

    set flags 0

    switch -exact -- $opts(type) {
        "any"       {set type 0}
        "disk"      {set type 1}
        "printer"   {set type 2}
        default {
            error "Invalid network share type '$opts(type)'"
        }
    }

    # localdevice - "" means no local device, * means pick any, otherwise
    # it's a local device to be mapped
    if {$opts(localdevice) == "*"} {
        set opts(localdevice) ""
        setbits flags 0x80;             # CONNECT_REDIRECT
    }

    if {$opts(defaultpassword) && $opts(nopassword)} {
        error "Options -defaultpassword and -nopassword may not be used together"
    }
    if {$opts(nopassword)} {
        set opts(password) ""
        set ignore_password 1
    } else {
        set ignore_password 0
        if {$opts(defaultpassword)} {
            set opts(password) ""
        }
    }

    foreach {opt mask} {
        interactive   0x8
        prompt        0x10
        updateprofile 0x1
        commandline   0x800
    } {
        if {$opts($opt)} {
            setbits flags $mask
        }
    }

    return [Twapi_WNetUseConnection $opts(window) $type $opts(localdevice) \
                $remoteshare $opts(provider) $opts(user) $ignore_password \
                $opts(password) $flags]
}

#
# Disconnects an existing share
proc twapi::disconnect_share {sharename args} {
    array set opts [cmdline::getoptions args {updateprofile force}]

    set flags [expr {$opts(updateprofile) ? 0x1 : 0}]
    WNetCancelConnection2 $sharename $flags $opts(force)
}

#
# Get connected share info
proc twapi::get_mapped_share_info {path args} {
    array set opts [cmdline::getoptions args {
        all user uncpath uncvolume relativepath
    }]

    if {! [regexp {^([[:alpha:]]:)} $path dontcare drive]} {
        error "No drive specified in path '$path'"
    }

    set result [list ]

    foreach {uncpath uncvolume relativepath} [WNetGetUniversalName $path] break
    foreach opt {uncpath uncvolume relativepath} {
        if {$opts($opt) || $opts(all)} {
            lappend result -$opt [set $opt]
        }
    }

    if {$opts(user) || $opts(all)} {
        lappend result -user [WNetGetUser $drive]
    }

    return $result
}


################################################################
# Utility functions

proc twapi::_share_type_symbols_to_code {typesyms} {
    variable windefs

    set special 0
    foreach sym $typesyms {
        switch -exact -- $sym {
            file    { set code $windefs(STYPE_DISKTREE) }
            printer { set code $windefs(STYPE_PRINTQ) }
            device  { set code $windefs(STYPE_DEVICE) }
            ipc     { set code $windefs(STYPE_IPC) }
            special { setbits special $windefs(STYPE_SPECIAL) }
            temporary { setbits special $windefs(STYPE_TEMPORARY) }
            default {
                error "Unknown type network share type symbol '$sym'"
            }
        }
    }

    if {![info exists code]} {
        error "No base type symbol specified for network share"
    }

    return [expr {$code | $special}]
}


# First element is always the base type of the share
proc twapi::_share_type_code_to_symbols {type} {
    variable windefs


    set special [expr {$type & ($windefs(STYPE_SPECIAL) | $windefs(STYPE_TEMPORARY))}]

    # We need the special cast to int because else operands get promoted
    # to 64 bits as the hex is treated as an unsigned value
    switch -exact -- [expr {int($type & ~ $special)}] \
        [list \
             $windefs(STYPE_DISKTREE) {set sym "file"} \
             $windefs(STYPE_PRINTQ)   {set sym "printer"} \
             $windefs(STYPE_DEVICE)   {set sym "device"} \
             $windefs(STYPE_IPC)      {set sym "ipc"} \
             default                  {set sym $type}
            ]

    set typesyms [list $sym]

    if {$special & $windefs(STYPE_SPECIAL)} {
        lappend typesyms special
    }

    if {$special & $windefs(STYPE_TEMPORARY)} {
        lappend typesyms temporary
    }
    
    return $typesyms
}